import numpy as np
import torch as th
import lightning as ltn

from torch.utils.data import DataLoader
from wxbtool.util.plotter import plot


class LightningModel(ltn.LightningModule):
    def __init__(self, model, opt=None):
        super(LightningModel, self).__init__()
        self.model = model
        self.learning_rate = 1e-3

        self.counter = 0
        self.labeled_loss = 0
        self.labeled_mse = 0

        self.opt = opt

        if opt and hasattr(opt, 'rate'):
            self.learning_rate = float(opt.rate)

    def configure_optimizers(self):
        optimizer = th.optim.Adam(self.parameters(), lr=self.learning_rate)
        scheduler = th.optim.lr_scheduler.CosineAnnealingLR(optimizer, 53)
        return [optimizer], [scheduler]

    def loss_fn(self, input, result, target):
        return self.model.lossfun(input, result, target)

    def compute_mse(self, targets, results):
        _, tgt = self.model.get_targets(**targets)
        _, rst = self.model.get_results(**results)
        tgt = (
            tgt.detach().cpu().numpy().reshape(-1, self.model.setting.pred_span, 32, 64)
        )
        rst = (
            rst.detach().cpu().numpy().reshape(-1, self.model.setting.pred_span, 32, 64)
        )
        mse = np.mean(self.model.weight.cpu().numpy() * (rst - tgt) ** 2)
        return mse

    def forecast_error(self, rmse):
        return rmse

    def forward(self, **inputs):
        return self.model(**inputs)

    def plot(self, inputs, results, targets):
        vars_in, _ = self.model.get_inputs(**inputs)
        for bas, var in enumerate(self.model.setting.vars_in):
            for ix in range(self.model.setting.input_span):
                img = vars_in[var][0, ix].detach().cpu().numpy().reshape(32, 64)
                plot(var, open("%s_inp_%d.png" % (var, ix), mode="wb"), img)

        vars_fc, _ = self.model.get_results(**results)
        vars_tg, _ = self.model.get_targets(**targets)
        for bas, var in enumerate(self.model.setting.vars_out):
            for ix in range(self.model.setting.pred_span):
                fcst = vars_fc[var][0, ix].detach().cpu().numpy().reshape(32, 64)
                tgrt = vars_tg[var][0, ix].detach().cpu().numpy().reshape(32, 64)
                plot(var, open("%s_fcs_%d.png" % (var, ix), mode="wb"), fcst)
                plot(var, open("%s_tgt_%d.png" % (var, ix), mode="wb"), tgrt)

    def training_step(self, batch, batch_idx):
        inputs, targets = batch
        inputs = {v: inputs[v].float() for v in self.model.setting.vars}
        targets = {v: targets[v].float() for v in self.model.setting.vars}
        results = self.forward(**inputs)

        loss = self.loss_fn(inputs, results, targets)

        self.log("train_loss", loss, prog_bar=True)
        return loss

    def validation_step(self, batch, batch_idx):
        inputs, targets = batch
        batch_len = inputs[self.model.setting.vars[0]].shape[0]

        inputs = {v: inputs[v].float() for v in self.model.setting.vars}
        targets = {v: targets[v].float() for v in self.model.setting.vars}
        results = self.forward(**inputs)
        loss = self.loss_fn(inputs, results, targets)
        mse = self.compute_mse(targets, results)

        self.labeled_loss += loss.item() * batch_len
        self.labeled_mse += mse * batch_len
        self.counter += batch_len

        rmse = np.sqrt(self.labeled_mse / self.counter)
        self.log("val_rmse", rmse, prog_bar=True)
        self.log("val_loss", loss, prog_bar=True)

        self.plot(inputs, results, targets)

    def test_step(self, batch, batch_idx):
        inputs, targets = batch
        batch_len = inputs[self.model.setting.vars[0]].shape[0]

        inputs = {v: inputs[v].float() for v in self.model.setting.vars}
        targets = {v: targets[v].float() for v in self.model.setting.vars}
        results = self.forward(**inputs)
        loss = self.loss_fn(inputs, results, targets)
        mse = self.compute_mse(targets, results)

        self.labeled_loss += loss.item() * batch_len
        self.labeled_mse += mse * batch_len
        self.counter += batch_len

        rmse = np.sqrt(self.labeled_mse / self.counter)
        self.log("test_loss", loss, prog_bar=True)
        self.log("test_rmse", rmse, prog_bar=True)

        self.plot(inputs, results, targets)

    def on_save_checkpoint(self, checkpoint) -> None:
        import glob
        import os

        rmse = np.sqrt(self.labeled_mse / self.counter)
        error = self.forecast_error(rmse)
        loss = self.labeled_loss / self.counter
        record = "%2.5f-%03d-%1.5f.ckpt" % (error, checkpoint["epoch"], loss)
        mname = self.model.name
        fname = f"trains/{mname}/best-{record}"
        os.makedirs(
            f"trains/{mname}", exist_ok=True
        )
        with open(fname, "bw") as f:
            th.save(checkpoint, f)
        for ix, ckpt in enumerate(sorted(glob.glob(f"trains/{mname}/best-*.ckpt"), reverse=False)):
            if ix > 5:
                os.unlink(ckpt)

        self.counter = 0
        self.labeled_loss = 0
        self.labeled_rmse = 0

        print()

    def train_dataloader(self):
        if self.model.dataset_train is None:
            if self.opt.data != "":
                self.model.load_dataset("train", "client", url=self.opt.data)
            else:
                self.model.load_dataset("train", "server")
        return DataLoader(
            self.model.dataset_train,
            batch_size=self.opt.batch_size,
            num_workers=self.opt.n_cpu,
            shuffle=True,
        )

    def val_dataloader(self):
        if self.model.dataset_eval is None:
            if self.opt.data != "":
                self.model.load_dataset("train", "client", url=self.opt.data)
            else:
                self.model.load_dataset("train", "server")
        return DataLoader(
            self.model.dataset_eval,
            batch_size=self.opt.batch_size,
            num_workers=self.opt.n_cpu,
            shuffle=True,
        )

    def test_dataloader(self):
        if self.model.dataset_test is None:
            if self.opt.data != "":
                self.model.load_dataset("train", "client", url=self.opt.data)
            else:
                self.model.load_dataset("train", "server")
        return DataLoader(
            self.model.dataset_test,
            batch_size=self.opt.batch_size,
            num_workers=self.opt.n_cpu,
            shuffle=True,
        )


class GANModel(LightningModel):
    def __init__(self, generator, discriminator, opt=None):
        super(GANModel, self).__init__(generator, opt=opt)
        self.generator = generator
        self.discriminator = discriminator
        self.learning_rate = 1e-4  # Adjusted for GANs
        self.automatic_optimization = False
        self.realness = 0
        self.fakeness = 1

        if opt and hasattr(opt, 'rate'):
            learning_rate = float(opt.rate)
            ratio = float(opt.ratio)
            self.generator.learning_rate = learning_rate
            self.discriminator.learning_rate = learning_rate / ratio

    def configure_optimizers(self):
        # Separate optimizers for generator and discriminator
        g_optimizer = th.optim.Adam(self.generator.parameters(), lr=self.generator.learning_rate)
        d_optimizer = th.optim.Adam(self.discriminator.parameters(), lr=self.discriminator.learning_rate)
        return [g_optimizer, d_optimizer]

    def generator_loss(self, fake_judgement):
        # Loss for generator (we want the discriminator to predict all generated images as real)
        return th.nn.functional.binary_cross_entropy_with_logits(fake_judgement["data"], th.ones_like(fake_judgement["data"], dtype=th.float32))

    def discriminator_loss(self, real_judgement, fake_judgement):
        # Loss for discriminator (real images should be classified as real, fake images as fake)
        real_loss = th.nn.functional.binary_cross_entropy_with_logits(real_judgement["data"], th.ones_like(real_judgement["data"], dtype=th.float32))
        fake_loss = th.nn.functional.binary_cross_entropy_with_logits(fake_judgement["data"], th.zeros_like(fake_judgement["data"], dtype=th.float32))
        return (real_loss + fake_loss) / 2

    def forecast_error(self, rmse):
        return self.generator.forecast_error(rmse)

    def training_step(self, batch, batch_idx):
        inputs, targets = batch
        inputs, _ = self.model.get_inputs(**inputs)
        targets, _ = self.model.get_targets(**targets)
        inputs['noise'] = th.randn_like(inputs['data'][:, :1, :, :], dtype=th.float32)

        g_optimizer, d_optimizer = self.optimizers()

        self.toggle_optimizer(g_optimizer)
        forecast = self.generator(**inputs)
        judgement = self.discriminator(**inputs, target=forecast["data"])
        forecast_loss = self.loss_fn(inputs, forecast, targets)
        generate_loss = self.generator_loss(judgement)
        total_loss = forecast_loss + generate_loss
        self.log("total", total_loss, prog_bar=True)
        self.log("forecast", forecast_loss, prog_bar=True)
        self.manual_backward(total_loss)
        g_optimizer.step()
        g_optimizer.zero_grad()
        self.untoggle_optimizer(g_optimizer)

        self.toggle_optimizer(d_optimizer)
        forecast = self.generator(**inputs)
        forecast["data"].detach()  # Detach to avoid generator gradient updates
        real_judgement = self.discriminator(**inputs, target=targets['data'])
        fake_judgement = self.discriminator(**inputs, target=forecast["data"])
        judgement_loss = self.discriminator_loss(real_judgement, fake_judgement)
        self.log("judgement", judgement_loss, prog_bar=True)
        self.manual_backward(judgement_loss)
        d_optimizer.step()
        d_optimizer.zero_grad()
        self.untoggle_optimizer(d_optimizer)

    def validation_step(self, batch, batch_idx):
        inputs, targets = batch
        inputs, _ = self.model.get_inputs(**inputs)
        targets, _ = self.model.get_targets(**targets)
        inputs['noise'] = th.randn_like(inputs['data'][:, :1, :, :], dtype=th.float32)
        forecast = self.generator(**inputs)
        forecast_loss = self.loss_fn(inputs, forecast, targets)
        real_judgement = self.discriminator(**inputs, target=targets['data'])
        fake_judgement = self.discriminator(**inputs, target=forecast["data"])
        self.realness = real_judgement["data"].mean().item()
        self.fakeness = fake_judgement["data"].mean().item()
        self.log("realness", self.realness, prog_bar=True)
        self.log("fakeness", self.fakeness, prog_bar=True)
        self.log("val_forecast", forecast_loss, prog_bar=True)
        self.log("val_loss", forecast_loss, prog_bar=True)

        mse = self.compute_mse(targets, forecast)
        batch_len = inputs['data'].shape[0]
        self.labeled_loss += forecast_loss.item() * batch_len
        self.labeled_mse += mse * batch_len
        self.counter += batch_len

        self.plot(inputs, forecast, targets)

    def test_step(self, batch, batch_idx):
        inputs, targets = batch
        inputs, _ = self.model.get_inputs(**inputs)
        targets, _ = self.model.get_targets(**targets)
        inputs['noise'] = th.randn_like(inputs['data'][:, :1, :, :], dtype=th.float32)
        forecast = self.generator(**inputs)
        forecast_loss = self.loss_fn(inputs, forecast, targets)
        real_judgement = self.discriminator(**inputs, target=targets['data'])
        fake_judgement = self.discriminator(**inputs, target=forecast["data"])
        self.realness = real_judgement["data"].mean().item()
        self.fakeness = fake_judgement["data"].mean().item()
        self.log("realness", self.realness, prog_bar=True)
        self.log("fakeness", self.fakeness, prog_bar=True)
        self.log("forecast", forecast_loss, prog_bar=True)
        self.plot(inputs, forecast, targets)

    def on_validation_epoch_end(self):
        balance = self.realness - self.fakeness
        self.log("balance", balance)
        if abs(balance - self.opt.balance) < self.opt.tolerance:
            self.trainer.should_stop = True



# -*- coding: utf-8 -*-

import importlib
import logging
import resource
import arrow
import sys
import os
import flask

import msgpack
import msgpack_numpy as m

m.patch()

import numpy as np  # noqa: E402

from pathlib import Path  # noqa: E402
from flask import Flask  # noqa: E402


rlimit = resource.getrlimit(resource.RLIMIT_NOFILE)
resource.setrlimit(resource.RLIMIT_NOFILE, (2048, rlimit[1]))
np.core.arrayprint._line_width = 150
np.set_printoptions(linewidth=np.inf)


datasets = {}


def init(opt):
    time_str = arrow.now().format("YYYYMMDD_HHmmss")
    model_path = Path(f"./dsserver/{time_str}")
    model_path.mkdir(exist_ok=True, parents=True)
    log_file = model_path / Path("dsserver.log")
    logging.basicConfig(level=logging.INFO, filename=log_file, filemode="w")
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    logger.info(str(opt))

    try:
        sys.path.insert(0, os.getcwd())
        mdm = importlib.import_module(opt.module, package=None)
        setting = getattr(mdm, opt.setting)()
        spec = getattr(mdm, "Spec")(setting)
    except ImportError:
        exc_info = sys.exc_info()
        print("failure when loading model")
        import traceback

        traceback.print_exception(*exc_info)
        del exc_info
        sys.exit(1)

    spec.load_dataset("train", "server")
    spec.load_dataset("test", "server")
    dtrain = spec.dataset_train
    deval = spec.dataset_eval
    dtest = spec.dataset_test
    datasets["train"] = dtrain
    datasets["eval"] = deval
    datasets["test"] = dtest

    handler = logging.StreamHandler(sys.stderr)
    handler.setFormatter(
        logging.Formatter("%(asctime)s %(levelname)s %(name)s: %(message)s")
    )
    logger.addHandler(handler)
    gunicorn_logger = logging.getLogger("gunicorn.info")
    app.logger.handlers.extend(gunicorn_logger.handlers)
    logger.handlers.extend(app.logger.handlers)


app = Flask(__name__)
app.debug = False

route = app.route


@route("/<string:hash>/<string:mode>")
def length(hash, mode):
    ds = datasets[mode]
    if ds.hashcode != hash:
        return flask.current_app.response_class(
            "not found", status=404, mimetype="application/msgpack"
        )

    app.logger.info("query length[%s] %d", mode, len(ds))
    msg = msgpack.dumps(
        {
            "size": len(ds),
        }
    )

    return flask.current_app.response_class(
        msg, status=200, mimetype="application/msgpack"
    )


@route("/<string:hash>/<string:mode>/<int:idx>")
def seek(hash, mode, idx):
    ds = datasets[mode]
    if ds.hashcode != hash:
        return flask.current_app.response_class(
            "not found", status=404, mimetype="application/msgpack"
        )

    app.logger.info("query data[%s] at %d", mode, idx)
    inputs, targets = ds[idx]
    msg = msgpack.dumps(
        {
            "inputs": inputs,
            "targets": targets,
        }
    )

    return flask.current_app.response_class(
        msg, status=200, mimetype="application/msgpack"
    )


def main(context, opt):
    import gunicorn.app.base

    init(opt)

    class StandaloneApplication(gunicorn.app.base.BaseApplication):
        def __init__(self, app, options=None):
            self.options = options or {}
            self.application = app
            super().__init__()

        def load_config(self):
            config = {
                key: value
                for key, value in self.options.items()
                if key in self.cfg.settings and value is not None
            }
            for key, value in config.items():
                self.cfg.set(key.lower(), value)

        def load(self):
            return self.application

    print("PID %s" % str(os.getpid()))
    print("serving... %s" % opt.module)
    print("bind: %s" % opt.bind)
    print("workers: %s" % opt.workers)

    if opt.test == "false":
        options = {
            "bind": opt.bind,
            "workers": opt.workers,
        }
        StandaloneApplication(app, options).run()



from arghandler import ArgumentHandler, subcmd

from wxbtool.data.download import main as dnmain
from wxbtool.data.dsserver import main as dsmain
from wxbtool.nn.train import main as tnmain
from wxbtool.nn.test import main as ttmain


@subcmd
def help(parser, context, args):
    pass


@subcmd("dserve", help="start the dataset server")
def dserve(parser, context, args):
    parser.add_argument(
        "-b",
        "--bind",
        type=str,
        default=None,
        help="binding address (ip:port or unix:/path/to/your.sock)",
    )
    parser.add_argument(
        "-i", "--ip", type=str, default="127.0.0.1", help="the ip of the dataset serevr"
    )
    parser.add_argument(
        "-p", "--port", type=int, default=8088, help="the port of the dataset serevr"
    )
    parser.add_argument(
        "-w", "--workers", type=int, default=4, help="the number of workers"
    )
    parser.add_argument(
        "-m",
        "--module",
        type=str,
        default="wxbtool.specs.res5_625.t850weyn",
        help="module of a metrological model to load",
    )
    parser.add_argument(
        "-s",
        "--setting",
        type=str,
        default="Setting",
        help="setting for a metrological model spec",
    )
    parser.add_argument(
        "-t", "--test", type=str, default="false", help="setting for test"
    )
    opt = parser.parse_args(args)

    dsmain(context, opt)


@subcmd("train", help="start training")
def train(parser, context, args):
    parser.add_argument("-g", "--gpu", type=str, default="0", help="index of gpu")
    parser.add_argument(
        "-c",
        "--n_cpu",
        type=int,
        default=8,
        help="number of cpu threads to use during batch generation",
    )
    parser.add_argument(
        "-b", "--batch_size", type=int, default=64, help="size of the batches"
    )
    parser.add_argument(
        "-e",
        "--epoch",
        type=int,
        default=0,
        help="current epoch to start training from",
    )
    parser.add_argument(
        "-n", "--n_epochs", type=int, default=200, help="number of epochs of training"
    )
    parser.add_argument(
        "-m",
        "--module",
        type=str,
        default="wxbtool.zoo.unet.t850d3",
        help="module of the metrological model to load",
    )
    parser.add_argument(
        "-l",
        "--load",
        type=str,
        default="",
        help="dump file of the metrological model to load",
    )
    parser.add_argument(
        "-k", "--check", type=str, default="", help="checkpoint file to load"
    )
    parser.add_argument(
        "-r", "--rate", type=str, default="0.001", help="learning rate"
    )
    parser.add_argument(
        "-w", "--weightdecay", type=float, default=0.0, help="weight decay"
    )
    parser.add_argument(
        "-R", "--ratio", type=str, default="10", help="the ratio of the two learning rates between generator and discriminator in GAN"
    )
    parser.add_argument(
        "-B", "--balance", type=float, default=0.9, help="exit balance for GAN training"
    )
    parser.add_argument(
        "-T", "--tolerance", type=float, default=0.05, help="exit balance tolerance for GAN training"
    )
    parser.add_argument(
        "-d",
        "--data",
        type=str,
        default="",
        help="http url of the dataset server or binding unix socket (unix:/path/to/your.sock)",
    )
    parser.add_argument(
        "-G", "--gan", type=str, default="false", help="training GAN or not, default is false"
    )
    parser.add_argument(
        "-t", "--test", type=str, default="false", help="setting for test"
    )

    opt = parser.parse_args(args)

    tnmain(context, opt)


@subcmd("test", help="start testing")
def test(parser, context, args):
    parser.add_argument("-g", "--gpu", type=str, default="0", help="index of gpu")
    parser.add_argument(
        "-c",
        "--n_cpu",
        type=int,
        default=8,
        help="number of cpu threads to use during batch generation",
    )
    parser.add_argument(
        "-b", "--batch_size",
        type=int,
        default=64,
        help="size of the batches"
    )
    parser.add_argument(
        "-m",
        "--module",
        type=str,
        default="wxbtool.zoo.unet.t850d3",
        help="module of the metrological model to load",
    )
    parser.add_argument(
        "-l",
        "--load",
        type=str,
        default="",
        help="dump file of the metrological model to load",
    )
    parser.add_argument(
        "-d",
        "--data",
        type=str,
        default="",
        help="http url of the dataset server or binding unix socket (unix:/path/to/your.sock)",
    )
    parser.add_argument(
        "-G", "--gan", type=str, default="false", help="training GAN or not, default is false"
    )
    parser.add_argument(
        "-t", "--test", type=str, default="false", help="setting for test"
    )
    opt = parser.parse_args(args)

    ttmain(context, opt)


@subcmd("download", help="download the latest hourly ERA5 data from ECMWF")
def download(parser, context, args):
    parser.add_argument(
        "-m",
        "--module",
        type=str,
        default="wxbtool.zoo.unet.t850d3",
        help="module of the metrological model to load",
    )
    parser.add_argument(
        "-G", "--gan", type=str, default="false", help="model is GAN or not, default is false"
    )
    parser.add_argument(
        "--coverage",
        type=str,
        choices=["daily", "weekly", "monthly"],
        default="weekly",
        help="specify the period for which data coverage is required",
    )
    parser.add_argument(
        "--retention",
        type=str,
        choices=["daily", "weekly", "monthly"],
        default="weekly",
        help="specify the retention period for keeping the latest data",
    )
    opt = parser.parse_args(args)

    dnmain(context, opt)


def main():
    import sys

    handler = ArgumentHandler()
    if len(sys.argv) < 2:
        handler.run(["help"])
    else:
        handler.run(sys.argv[1:])



import os
import sys
import importlib

import torch as th
import lightning.pytorch as pl

from lightning.pytorch.callbacks import EarlyStopping
from wxbtool.nn.lightning import LightningModel, GANModel


if th.cuda.is_available():
    accelerator = "gpu"
    th.set_float32_matmul_precision("medium")
elif th.backends.mps.is_available():
    accelerator = "cpu"
else:
    accelerator = "cpu"


def main(context, opt):
    os.environ["CUDA_VISIBLE_DEVICES"] = opt.gpu
    try:
        sys.path.insert(0, os.getcwd())
        mdm = importlib.import_module(opt.module, package=None)

        if opt.gan == "true":
            learning_rate = float(opt.rate)
            ratio = float(opt.ratio)
            generator_lr, discriminator_lr = learning_rate, learning_rate / ratio
            model = GANModel(mdm.generator, mdm.discriminator, opt=opt)
            model.generator.learning_rate = generator_lr
            model.discriminator.learning_rate = discriminator_lr
            callbacks = []
        else:
            learning_rate = float(opt.rate)
            model = LightningModel(mdm.model, opt=opt)
            model.learning_rate = learning_rate
            callbacks = [EarlyStopping(monitor="val_loss", mode="min", patience=30)]

        n_epochs = 1 if opt.test == "true" else opt.n_epochs

        trainer = pl.Trainer(
            accelerator=accelerator,
            precision=32,
            max_epochs=n_epochs,
            callbacks=callbacks,
        )

        if opt.load:
            model.load_from_checkpoint(opt.load)

        trainer.fit(model)
        trainer.test(model=model, dataloaders=model.test_dataloader())
    except ImportError as e:
        exc_info = sys.exc_info()
        print(e)
        print("failure when training model")
        import traceback

        traceback.print_exception(*exc_info)
        del exc_info
        sys.exit(-1)



# -*- coding: utf-8 -*-

vars2d = [
    "2m_temperature",
    "10m_u_component_of_wind",
    "10m_v_component_of_wind",
    "total_cloud_cover",
    "total_precipitation",
    "toa_incident_solar_radiation",
    "temperature_850hPa",
]

vars3d = [
    "geopotential",
    "temperature",
    "specific_humidity",
    "relative_humidity",
    "u_component_of_wind",
    "v_component_of_wind",
    "vorticity",
    "potential_vorticity",
]

codes = {
    "geopotential": "z",
    "temperature": "t",
    "temperature_850hPa": "t",
    "specific_humidity": "q",
    "relative_humidity": "r",
    "u_component_of_wind": "u",
    "v_component_of_wind": "v",
    "vorticity": "vo",
    "potential_vorticity": "pv",
    "2m_temperature": "t2m",
    "10m_u_component_of_wind": "u10",
    "10m_v_component_of_wind": "v10",
    "total_cloud_cover": "tcc",
    "total_precipitation": "tp",
    "toa_incident_solar_radiation": "tisr",
}

code2var = {
    "z": "geopotential",
    "t": "temperature",
    "q": "specific_humidity",
    "r": "relative_humidity",
    "u": "u_component_of_wind",
    "v": "v_component_of_wind",
    "vo": "vorticity",
    "pv": "potential_vorticity",
    "t2m": "2m_temperature",
    "u10": "10m_u_component_of_wind",
    "v10": "10m_v_component_of_wind",
    "tcc": "total_cloud_cover",
    "tp": "total_precipitation",
    "tisr": "toa_incident_solar_radiation",
}


def split_name(composite):
    if (
        composite == "t2m"
        or composite == "u10"
        or composite == "v10"
        or composite == "tcc"
        or composite == "tp"
        or composite == "tisr"
    ):
        return composite, ""
    else:
        if composite[:2] == "vo" or composite[:2] == "pv":
            return composite[:2], composite[2:]
        else:
            return composite[:1], composite[1:]



import os
import shutil
import sys
import datetime
import importlib
import logging
import cdsapi

from dataclasses import dataclass, field
from typing import List, Optional

from wxbtool.nn.lightning import GANModel, LightningModel

# Configure logging to display information and errors
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(levelname)s - %(message)s',
    handlers=[
        logging.StreamHandler()
    ]
)


@dataclass
class Config:
    output_folder: str
    variables: List[str]
    vars2d: List[str]
    vars3d: List[str]
    levels: List[str]
    coverage: str  # 'daily', 'weekly', 'monthly'
    retention_days: int = 7  # Number of days to retain data
    retention_action: str = "delete"  # Options: "delete", "archive"
    archive_folder: Optional[str] = None  # Required if action is "archive"
    reference_time_delta: datetime.timedelta = field(default_factory=lambda: datetime.timedelta(weeks=1))
    grid: List[float] = field(default_factory=lambda: [5.625, 5.625])  # Spatial resolution
    area: List[float] = field(default_factory=lambda: [90, -180, -90, 180])  # Global coverage


class ERA5Downloader:
    def __init__(self, config: Config):
        self.c = cdsapi.Client()
        self.config = config
        self.ensure_output_dirs()

    def ensure_output_dirs(self):
        """Create the base directories for each variable."""
        for variable in self.config.variables:
            var_path = os.path.join(self.config.output_folder, variable)
            os.makedirs(var_path, exist_ok=True)
        logging.info("Base output directories are set up.")
        # If archiving is enabled, ensure the archive folder exists
        if self.config.retention_action == "archive" and self.config.archive_folder:
            os.makedirs(self.config.archive_folder, exist_ok=True)
            logging.info("Archive directory is set up.")

    def get_time_span(self):
        """Determine the start and end dates based on the coverage type."""
        end_date = datetime.datetime.utcnow() - self.config.reference_time_delta
        if self.config.coverage == "daily":
            start_date = end_date - datetime.timedelta(days=1)
        elif self.config.coverage == "weekly":
            start_date = end_date - datetime.timedelta(weeks=1)
        elif self.config.coverage == "monthly":
            start_date = end_date - datetime.timedelta(days=30)  # Approximate month as 30 days
        else:
            raise ValueError("Unsupported coverage type. Use 'daily', 'weekly', or 'monthly'.")
        return start_date, end_date

    def generate_datetime_list(self, start_date: datetime.datetime, end_date: datetime.datetime):
        """Generate a list of (date, hour) tuples for each hour within the time span."""
        current_datetime = start_date.replace(hour=0, minute=0, second=0, microsecond=0)
        end_datetime = end_date.replace(hour=23, minute=0, second=0, microsecond=0)
        while current_datetime <= end_datetime:
            yield current_datetime, f"{current_datetime.hour:02}:00"
            current_datetime += datetime.timedelta(hours=1)

    def build_filename(self, variable: str, date: datetime.datetime, time_str: str) -> str:
        """Construct the filename for the NetCDF file."""
        year = date.strftime("%Y")
        month = date.strftime("%m")
        day = date.strftime("%d")
        hour = time_str.split(":")[0]
        filename = f"{date.strftime('%Y%m%d')}_{hour}.nc"
        return os.path.join(
            self.config.output_folder,
            variable,
            year,
            month,
            filename
        )

    def ensure_variable_dirs(self, variable: str, date: datetime.datetime):
        """Ensure that the directory for a specific variable, year, and month exists."""
        year = date.strftime("%Y")
        month = date.strftime("%m")
        var_year_path = os.path.join(self.config.output_folder, variable, year)
        var_month_path = os.path.join(var_year_path, month)
        os.makedirs(var_month_path, exist_ok=True)

    def retrieve_data(self, variable: str, date: datetime.datetime, time_str: str, filename: str):
        """Retrieve data for a specific variable, date, and time."""
        request_params = {
            "product_type": "reanalysis",
            "variable": variable,
            "year": date.strftime("%Y"),
            "month": date.strftime("%m"),
            "day": date.strftime("%d"),
            "time": time_str,
            "format": "netcdf",
            "grid": self.config.grid,  # Spatial resolution
            "area": self.config.area,  # Global coverage
        }

        if variable in self.config.vars3d:
            request_params["pressure_level"] = self.config.levels
            dataset = "reanalysis-era5-pressure-levels"
        elif variable in self.config.vars2d:
            dataset = "reanalysis-era5-single-levels"
        else:
            logging.warning(f"Variable '{variable}' is not recognized as single or pressure level.")
            return

        try:
            self.c.retrieve(dataset, request_params, filename)
            logging.info(f"Downloaded: {filename}")
        except Exception as e:
            logging.error(f"Failed to download {filename}: {e}")

    def manage_retention(self):
        """Manage data retention by deleting or archiving old files."""
        retention_threshold = datetime.datetime.utcnow() - datetime.timedelta(days=self.config.retention_days)
        logging.info(f"Managing retention: deleting or archiving files older than {self.config.retention_days} days.")

        for variable in self.config.variables:
            var_path = os.path.join(self.config.output_folder, variable)
            for root, dirs, files in os.walk(var_path):
                for file in files:
                    if file.endswith(".nc"):
                        file_path = os.path.join(root, file)
                        try:
                            file_mtime = datetime.datetime.utcfromtimestamp(os.path.getmtime(file_path))
                            if file_mtime < retention_threshold:
                                if self.config.retention_action == "delete":
                                    os.remove(file_path)
                                    logging.info(f"Deleted old file: {file_path}")
                                elif self.config.retention_action == "archive" and self.config.archive_folder:
                                    # Determine archive subdirectory based on variable, year, month
                                    relative_path = os.path.relpath(root, self.config.output_folder)
                                    archive_dir = os.path.join(self.config.archive_folder, relative_path)
                                    os.makedirs(archive_dir, exist_ok=True)
                                    shutil.move(file_path, os.path.join(archive_dir, file))
                                    logging.info(f"Archived old file: {file_path} to {archive_dir}")
                        except Exception as e:
                            logging.error(f"Error processing file {file_path}: {e}")

    def download(self):
        """Execute the download process based on the configuration."""
        start_date, end_date = self.get_time_span()
        logging.info(f"Downloading data from {start_date.strftime('%Y-%m-%d %H:%M')} to {end_date.strftime('%Y-%m-%d %H:%M')}.")

        for variable in self.config.variables:
            for date, time_str in self.generate_datetime_list(start_date, end_date):
                self.ensure_variable_dirs(variable, date)
                filename = self.build_filename(variable, date, time_str)
                if not os.path.exists(filename):
                    self.retrieve_data(variable, date, time_str, filename)
                else:
                    logging.info(f"File already exists: {filename}")

        # After downloading, manage retention
        self.manage_retention()


def main(context, opt):
    output_folder = "era5"
    try:
        sys.path.insert(0, os.getcwd())
        mdm = importlib.import_module(opt.module, package=None)
        if opt.gan == "true":
            model = GANModel(mdm.generator, mdm.discriminator, opt=opt)
        else:
            model = LightningModel(mdm.model, opt=opt)
        setting = model.model.setting
        resolution = float(setting.resolution.replace("deg", ""))
        variables = setting.vars
        vars2d = setting.vars2d
        vars3d = setting.vars3d
        levels = setting.levels

        # check .cdsapirc file in the home directory, if not exist, create one by prompting user to input the key
        if not os.path.exists(os.path.expanduser("~/.cdsapirc")):
            key = input("Please enter your ECMWF API key: ")
            with open(os.path.expanduser("~/.cdsapirc"), "w") as f:
                f.write("url: https://cds.climate.copernicus.eu/api/v2\n")
                f.write(f"key: {key}\n")

        config = Config(
            output_folder="era5",  # Replace with your desired output path
            variables=variables,
            vars2d=vars2d,
            vars3d=vars3d,
            levels=levels,
            grid=[resolution, resolution],  # Spatial resolution
            area=[90, -180, -90, 180],  # Global coverage
            coverage = "weekly",  # Options: "daily", "weekly", "monthly"
            retention_days = {"daily": 1, "weekly": 7, "monthly": 30}[opt.retention],
            retention_action = "delete",  # Options: "delete", "archive"
        )

        # Initialize and run the downloader
        downloader = ERA5Downloader(config)
        downloader.download()
    except Exception as e:
        exc_info = sys.exc_info()
        print(e)
        print("failure when downloading data")
        import traceback

        traceback.print_exception(*exc_info)
        del exc_info
        sys.exit(-1)
